#pragma once

#include <string>

#include <dune/common/timer.hh>

// AMDiS includes
#include <amdis/CreatorInterface.hpp>
#include <amdis/Output.hpp>
#include <amdis/linearalgebra/LinearSolverInterface.hpp>

namespace AMDiS
{
  /** \ingroup Solver
   *
   * \brief
   * Wrapper class for various solvers. These solvers
   * are parametrized by MatrixType and VectorType. The different
   * algorithms, like krylov subspace methods or other external
   * solvers where the backend provides an interface, can be assigned
   * by different Runner objects.
   **/
  template <class Traits, class Runner>
  class LinearSolver
      : public LinearSolverInterface<Traits>
  {
    using Self = LinearSolver;
    using Super = LinearSolverInterface<Traits>;

    using RunnerBase = typename Super::RunnerBase;
    using Mat = typename Traits::Mat;
    using Sol = typename Traits::Sol;
    using Rhs = typename Traits::Rhs;
    using Comm = typename Traits::Comm;

  public:
    /// A creator to be used instead of the constructor.
    struct Creator : CreatorInterfaceName<Super>
    {
      std::unique_ptr<Super> createWithString(std::string prefix) override
      {
        return std::make_unique<Self>(prefix);
      }
    };

  public:
    /// Constructor
    explicit LinearSolver(std::string prefix)
      : runner_(std::make_shared<Runner>(prefix))
    {}

    /// Implements \ref LinearSolverInterface::runner()
    virtual std::shared_ptr<RunnerBase> runner() override
    {
      return runner_;
    }

  private:
    /// Implements \ref LinearSolverInterface::solveSystemImpl()
    void solveImpl(Mat const& A, Sol& x, Rhs const& b, Comm& comm, SolverInfo& solverInfo) override
    {
      Dune::Timer t;
      if (solverInfo.doCreateMatrixData()) {
        // init matrix or wrap block-matrix or ...
        runner_->init(A, comm);
      }

      if (solverInfo.info() > 0)
        msg("fill matrix needed {} seconds", t.elapsed());

      int error = runner_->solve(A, x, b, solverInfo);
      solverInfo.setError(error);

      if (!solverInfo.doStoreMatrixData())
        runner_->exit();
    }

  private:
    /// redirect the implementation to a runner. Implements a \ref RunnerInterface
    std::shared_ptr<Runner> runner_;
  };

} // end namespace AMDiS
